#include "function.h"
#include <iostream>
#include <fstream>
#include <iomanip>
#include <algorithm>
#include <ranges>

namespace rv = std::ranges::views;


subject::subject(
	const char* s_sub_name,
	const char* s_teacher_name,
	int s_sub_id,
	int s_group_id,
	int s_num_of_st,
	bool s_availability,
	const char* s_type_control)
	: sub_id(s_sub_id)
	, group_id(s_group_id)
	, num_of_st(s_num_of_st)
	, availability(s_availability)
{
	strcpy_s(sub_name, 128, s_sub_name);
	strcpy_s(teacher_name, 128, s_teacher_name);
	strcpy_s(type_control, 8, s_type_control);
	if (strcmp(s_type_control, "zachot") == 0) {
		lecture_hours = s_num_of_st * 0.35;
		practice_hours = s_num_of_st * 0.35;
	}
	else if (strcmp(s_type_control, "exam") == 0) {
		lecture_hours = s_num_of_st * 0.5;
		practice_hours = s_num_of_st * 0.5;
	}
}

void subject::GetHours(subject& subj) {
	if (strcmp(subj.type_control, "zachot") == 0) {
		subj.lecture_hours = subj.num_of_st * 0.35;
		subj.practice_hours = subj.num_of_st * 0.35;
	}
	else if (strcmp(subj.type_control, "exam") == 0) {
		subj.lecture_hours = subj.num_of_st * 0.5;
		subj.practice_hours = subj.num_of_st * 0.5;
	}
}

std::istream& operator >> (std::istream& s_in, subject& subj)
{
	
	s_in.getline(subj.sub_name, 128);
	s_in.getline(subj.teacher_name, 128);
	s_in >> subj.sub_id
		>> subj.group_id
		>> subj.num_of_st
		>> subj.availability;
	s_in.ignore(std::numeric_limits<std::streamsize>::max(), '\n');
	s_in.getline(subj.type_control, 8);
	subject::GetHours(subj);	
	return s_in;
}

std::ostream& operator << (std::ostream& s_out, const subject& subj)
{
	s_out << subj.sub_name << std::endl
		<< subj.teacher_name << std::endl
		<< subj.sub_id << std::endl
		<< subj.group_id << std::endl
		<< subj.num_of_st << std::endl
		<< subj.lecture_hours << std::endl
		<< subj.practice_hours << std::endl
		<< subj.availability << std::endl
		<< subj.type_control << std::endl;
	return s_out;
}



function::function(const std::string& s_filename)
	:filename(s_filename)
{}

void function::CreateFile()
{
	if (first_create)
		std::cout << "FILE IS EXISTED" << std::endl;
	else {
		std::ofstream fout(filename, std::ios::binary);
		int stw, count = 0;
		subject el;
		std::cout << "Input information of new subjects" << std::endl;
		std::cout << "| Subject | Teacher | Subject ID | Group ID | Number of students | Availability | Type of final control |" << std::endl;
		do {
			if(count!=0) std::cin.ignore(std::numeric_limits<std::streamsize>::max(), '\n');
			std::cin >> el;
			fout.write(reinterpret_cast<char*>(&el), sizeof(subject));
			std::cout << "0 - Stop writing" << std::endl
				<< "1 - Continue writing" << std::endl
				<< ">> ";
			count++;
			std::cin >> stw;
			if (stw < 0 || stw >1) {
				std::cout << "Unknown command" << std::endl;
				break;
			}
		} while (stw == 1);
	}
}

void function::GetData()
{
	std::ifstream fin(filename, std::ios::binary);
	subject el;
	int idx = 0;
	while (fin.read(reinterpret_cast<char*>(&el), sizeof(subject))) {
		idx++;
		m_arr.insert(std::make_pair(idx, el));
	}
}

void function::Add(const std::initializer_list<subject>& s_list)
{
	std::ofstream fout(filename, std::ios::binary | std::ios::app);
	for (const auto& el : s_list)
		fout.write(reinterpret_cast<const char*>(&el), sizeof(subject));
}

void function::Delete(int s_idx)
{
	GetData();
	m_arr.erase(s_idx);
	std::ofstream fout(filename + " (updated)", std::ios::binary);
	for (const auto& el : m_arr)
		fout.write(reinterpret_cast<const char*>(&el.second), sizeof(subject));
}

void function::Replace(int s_idx, const subject& s_subj)
{
	GetData();
	std::ofstream fout(filename + " (updated)", std::ios::binary);
	m_arr[s_idx] = s_subj;
	for (const auto& el : m_arr)
		fout.write(reinterpret_cast<const char*>(&el.second), sizeof(subject));
}

void function::TransformToText() {
	GetData();
	std::ofstream fout(filename + ".txt");
	fout << "--------------------------------------------------------------------------- Subjects' status --------------------------------------------------------------------------" << std::endl;
	fout << "| N |       Subject       |        Teacher       | Subject ID | Group ID | Number of students | Lecture hours | Practice hours | Availability | Type of final control |" << std::endl;
	size_t count = 0;
	for (const auto& el : m_arr) {
		count++;
		fout << "|" << std::setw(3) << count << "|"
			<< std::setw(21) << el.second.sub_name << "|"
			<< std::setw(22) << el.second.teacher_name << "|"
			<< std::setw(12) << el.second.sub_id << "|"
			<< std::setw(10) << el.second.group_id << "|"
			<< std::setw(20) << el.second.num_of_st << "|"
			<< std::setw(15) << el.second.lecture_hours << "|"
			<< std::setw(16) << el.second.practice_hours << "|"
			<< std::setw(14) << el.second.availability << "|"
			<< std::setw(23) << el.second.type_control << "|"
			<< std::endl;
	}
}

void function::Search_by_teacher(const std::string& s_teacher)
{
	GetData();
	std::ofstream fout(s_teacher, std::ios::binary);
	std::vector<subject> temp;
	std::ranges::copy(
		m_arr
		| rv::transform([](const std::pair<int, subject>& el) {return el.second; })
                // ПОЧЕМУ s_teacher ПЕРЕДАЁТСЯ ПО ССЫЛКЕ?
                // ВЕДЬ ЭТА ПЕРЕМЕННАЯ НЕ ИЗМЕНЯЕТСЯ ВНУТРИ ЛЯМБДА ФУНКЦИИ
		| rv::filter([](const subject& el) {return el.teacher_name == s_teacher; })
		, std::back_inserter(temp));
	for (const auto& el : temp)
		fout.write(reinterpret_cast<const char*>(&el), sizeof(subject));
}

void function::Search_by_type_control(const std::string& s_type)
{
	GetData();
	std::ofstream fout(s_type, std::ios::binary);
	std::vector<subject> temp;
	std::ranges::copy(
		m_arr
		| rv::transform([](const std::pair<int, subject>& el) {return el.second; })
		| rv::filter([&s_type](const subject& el) {return el.type_control == s_type; })
		, std::back_inserter(temp));
	for (const auto& el : temp)
		fout.write(reinterpret_cast<const char*>(&el), sizeof(subject));
}

void function::Search_by_group_id(int s_id)
{
	GetData();
	std::ofstream fout(std::to_string(s_id), std::ios::binary);
	std::vector<subject> temp;
	std::ranges::copy(
		m_arr
		| rv::transform([](const std::pair<int, subject>& el) {return el.second; })
		| rv::filter([&s_id](const subject& el) {return el.group_id == s_id; })
		, std::back_inserter(temp));
	for (const auto& el : temp)
		fout.write(reinterpret_cast<const char*>(&el), sizeof(subject));
}

void function::Search_by_availability(bool s_avail)
{
	GetData();
	std::ofstream fout(std::to_string(s_avail), std::ios::binary);
	std::vector<subject> temp;
	std::ranges::copy(
		m_arr
		| rv::transform([](const std::pair<int, subject>& el) {return el.second; })
		| rv::filter([&s_avail](const subject& el) {return el.availability == s_avail; })
		, std::back_inserter(temp));
	for (const auto& el : temp)
		fout.write(reinterpret_cast<const char*>(&el), sizeof(subject));
}
